/*
yasep/C/yhx2bin.c
  This file is distributed under terms of the Affero GPL v3 license or later, see http://yasep.org
created mer. févr. 15 18:59:04 CET 2012 by whygee@f-cpu.org
version jeu. févr. 16 09:39:53 CET 2012

Import a .yhx (Yasep HeX) file and transform it to raw binary.
main() is also provided for testing and standalone operation.

test command :
gcc -DTEST_YHX -Wall -Os yhx2bin.c -o yhx2bin &&\
  ./yhx2bin example.yhx example.bin |\
     hexdump -e '"%06.6_ax: " 8/1 "%02X " "\n"' &&\
       diff example.bin example.ref.bin

Syntax of a YHX file :
 * first bytes : signature + new line
 * comments start with ";"
 * data are single bytes in sequential order,
    byte at address 0 comes first.
    They are represented by an atomic pair of hexa digits.
 * Spaces (non digit characters like \t, \n, whatever)
    can come anywhere in the stream,
    as long as it's not in the middle of a byte.
 * Address specifiers consist of a ":" followed by
    as many hexa digits for the address, followed by a space.
    the address is not checked because it is condidered
    to be a byte stream, but the feature is reserved for later.
 * default start address : 0

The methods are provided to the main routine with function pointers,
you have to provide your own, for example if you handle memory arrays.
The syntax is such that it is read linearly, not more than one
character at a time.

Afterthought : maybe a flex syntax would have been better
but i'm a bit rusty and I prefer to avoid complex tools.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

char YHX_signature[6]=";;yhx0";
int digit, nb_bytes_out;

#ifdef TEST_YHX
FILE *f_in, *f_out;

void error(char *msg, int nr, char *s) {
  fprintf(stderr,"\nyhx2bin: exiting on error #%i : %s\n%s\n",nr,msg,s);//strerror(errno));
  exit(EXIT_FAILURE);
}

void test_stop() {
  switch(digit) {
    case 0: break; // OK
    case 1:
    case 2:  error("unterminated byte",7,"");
    case -1: error("incomplete signature",8, "");
    default: error("something terribly wrong is going on",9,"O_0");
  }
  if (fflush(f_out) != 0)
    error("can't flush output file",10,strerror(errno));
  fprintf(stderr,"\n%u bytes written.\n",nb_bytes_out);
  exit(EXIT_SUCCESS);
}

char test_get() {
  int i=fgetc(f_in);
  if (i==EOF)
    test_stop();
  return (char)i;
}

void test_put(char c) {
  fprintf(stderr,"%02X ",c&255);
  nb_bytes_out++;
  fputc(c, f_out);
}

void test_seek(int a) {
  fprintf(stderr,"\njumping to address %X\n",a);
}
#else
 // to be done later
#endif

void read_yhx(char (*get)(), void (*put)(char), void (*seek)(int),  void (*stop)(void)) {
  int i,j=0;
  char c;

  void hex_digit(int c) {
    j=(j<<4) | c;
    if(digit++ ==1) {
      put(j);
      j=0;
      digit=0;
    }
  }

  // check the header
  digit=-1;
  for (i=0; i<sizeof(YHX_signature); i++) {
    if (get() != YHX_signature[i])
      error("Signature mismatch",4,"");
  }
  c=get();
  if (c!=0xD && c!=0xA)
    error("Signature mismatch",5,"");
  while( c==0xD || c==0xA )
    c=get(); // get to the next line.

  // OK this header is valid. proceed to parse the bytes.
  digit=0;
  nb_bytes_out=0;
  while(1) {
//    if (c==':') { // address specifier
      // TBD
//      c=get();
//    }
//    else {
      if (c==';') {
        while( c!=0xD && c!=0xA )
          c=get(); // swallow the line
        while( c==0xD || c==0xA )
          c=get(); // get to the next line.
      }
      else {
        if ( ((c>='0') && (c<='9')) ) // digit !
          hex_digit(c-'0');
        else
          if ( ((c>='A') && (c<='F')) )
            hex_digit(c-'A'+10);
          else
            if ( ((c>='a') && (c<='f')) )
              hex_digit(c-'a'+10);
//          else (fold)
              c=get(); // dumb char. space, line feed... get next one !
      }
    //}
  }
}


#ifdef TEST_YHX

int main(int argc, char **argv) {
  if (argc != 3)
    error("\nUsage : yhx2bin (filename in) (filename out)\nYou may use '-' to specify stdin and stdout.",1,"");

  // input file
  f_in=stdin;
  if (argv[1][0]!='-' || argv[1][1]!=0) {
    f_in=fopen(argv[1],"rb");
    if(f_in==NULL)
      error("Can't open input file for reading.",2,strerror(errno));
  }
  else
    fprintf(stderr,"Reading YHX from stdin\n");

  // output file
  f_out=stdout;
  if (argv[2][0]!='-' || argv[2][1]!=0) {
    f_out=fopen(argv[2],"wb");
    if(f_out==NULL)
      error("Can't open output file for writing.",3,strerror(errno));
  }
  else
    fprintf(stderr,"Writing binary to stdout\n");

  read_yhx(test_get, test_put, test_seek, test_stop);
  return 0;
}
#endif